# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

apply_delayed_binary_operation <- function(left_input, right_input, op) {
    .Call('_beachmat_apply_delayed_binary_operation', PACKAGE = 'beachmat', left_input, right_input, op)
}

apply_delayed_log <- function(raw_input, base) {
    .Call('_beachmat_apply_delayed_log', PACKAGE = 'beachmat', raw_input, base)
}

apply_delayed_unary_math <- function(raw_input, op) {
    .Call('_beachmat_apply_delayed_unary_math', PACKAGE = 'beachmat', raw_input, op)
}

apply_delayed_round <- function(raw_input) {
    .Call('_beachmat_apply_delayed_round', PACKAGE = 'beachmat', raw_input)
}

apply_delayed_associative_arithmetic <- function(raw_input, val, row, op) {
    .Call('_beachmat_apply_delayed_associative_arithmetic', PACKAGE = 'beachmat', raw_input, val, row, op)
}

apply_delayed_nonassociative_arithmetic <- function(raw_input, val, right, row, op) {
    .Call('_beachmat_apply_delayed_nonassociative_arithmetic', PACKAGE = 'beachmat', raw_input, val, right, row, op)
}

apply_delayed_comparison <- function(raw_input, val, row, op) {
    .Call('_beachmat_apply_delayed_comparison', PACKAGE = 'beachmat', raw_input, val, row, op)
}

apply_delayed_boolean <- function(raw_input, val, row, op) {
    .Call('_beachmat_apply_delayed_boolean', PACKAGE = 'beachmat', raw_input, val, row, op)
}

apply_delayed_boolean_not <- function(raw_input) {
    .Call('_beachmat_apply_delayed_boolean_not', PACKAGE = 'beachmat', raw_input)
}

apply_delayed_subset <- function(raw_input, subset, row) {
    .Call('_beachmat_apply_delayed_subset', PACKAGE = 'beachmat', raw_input, subset, row)
}

apply_delayed_transpose <- function(raw_input) {
    .Call('_beachmat_apply_delayed_transpose', PACKAGE = 'beachmat', raw_input)
}

apply_delayed_bind <- function(input, row) {
    .Call('_beachmat_apply_delayed_bind', PACKAGE = 'beachmat', input, row)
}

initialize_dense_matrix <- function(raw_x, nrow, ncol) {
    .Call('_beachmat_initialize_dense_matrix', PACKAGE = 'beachmat', raw_x, nrow, ncol)
}

initialize_dense_matrix_from_vector <- function(raw_x, nrow, ncol) {
    .Call('_beachmat_initialize_dense_matrix_from_vector', PACKAGE = 'beachmat', raw_x, nrow, ncol)
}

fragment_sparse_rows <- function(i, p, limits) {
    .Call('_beachmat_fragment_sparse_rows', PACKAGE = 'beachmat', i, p, limits)
}

sparse_subset_index <- function(starts, newp) {
    .Call('_beachmat_sparse_subset_index', PACKAGE = 'beachmat', starts, newp)
}

initialize_sparse_matrix <- function(raw_x, raw_i, raw_p, nrow, ncol, byrow) {
    .Call('_beachmat_initialize_sparse_matrix', PACKAGE = 'beachmat', raw_x, raw_i, raw_p, nrow, ncol, byrow)
}

initialize_SVT_SparseMatrix <- function(nr, nc, seed) {
    .Call('_beachmat_initialize_SVT_SparseMatrix', PACKAGE = 'beachmat', nr, nc, seed)
}

tatami_dim <- function(raw_input) {
    .Call('_beachmat_tatami_dim', PACKAGE = 'beachmat', raw_input)
}

tatami_is_sparse <- function(raw_input) {
    .Call('_beachmat_tatami_is_sparse', PACKAGE = 'beachmat', raw_input)
}

tatami_prefer_rows <- function(raw_input) {
    .Call('_beachmat_tatami_prefer_rows', PACKAGE = 'beachmat', raw_input)
}

tatami_column <- function(raw_input, i) {
    .Call('_beachmat_tatami_column', PACKAGE = 'beachmat', raw_input, i)
}

tatami_row <- function(raw_input, i) {
    .Call('_beachmat_tatami_row', PACKAGE = 'beachmat', raw_input, i)
}

tatami_row_sums <- function(raw_input, threads) {
    .Call('_beachmat_tatami_row_sums', PACKAGE = 'beachmat', raw_input, threads)
}

tatami_column_sums <- function(raw_input, threads) {
    .Call('_beachmat_tatami_column_sums', PACKAGE = 'beachmat', raw_input, threads)
}

tatami_realize <- function(raw_input, threads) {
    .Call('_beachmat_tatami_realize', PACKAGE = 'beachmat', raw_input, threads)
}

tatami_multiply_vector <- function(raw_input, other, right, num_threads) {
    .Call('_beachmat_tatami_multiply_vector', PACKAGE = 'beachmat', raw_input, other, right, num_threads)
}

tatami_multiply_columns <- function(raw_input, other, right, num_threads) {
    .Call('_beachmat_tatami_multiply_columns', PACKAGE = 'beachmat', raw_input, other, right, num_threads)
}

tatami_multiply_matrix <- function(raw_input, more_input, right, num_threads) {
    .Call('_beachmat_tatami_multiply_matrix', PACKAGE = 'beachmat', raw_input, more_input, right, num_threads)
}

initialize_unknown_matrix <- function(input) {
    .Call('_beachmat_initialize_unknown_matrix', PACKAGE = 'beachmat', input)
}

